@@ -3,21 +3,20 @@
 oauthlib.oauth2.rfc6749.grant_types.openid_connect
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 """
-from __future__ import unicode_literals, absolute_import
-
-from json import loads
-import logging
+from __future__ import absolute_import, unicode_literals
 
 import datetime
+import logging
+from json import loads
 
-from .base import GrantTypeBase
+from ..errors import ConsentRequired, InvalidRequestError, LoginRequired
+from ..request_validator import RequestValidator
 from .authorization_code import AuthorizationCodeGrant
 from .implicit import ImplicitGrant
-from ..errors import InvalidRequestError, LoginRequired, ConsentRequired
-from ..request_validator import RequestValidator
 
 log = logging.getLogger(__name__)
 
+
 class OIDCNoPrompt(Exception):
     """Exception used to inform users that no explicit authorization is needed.
 
@@ -64,7 +63,7 @@ class AuthCodeGrantDispatcher(object):
     def _handler_for_request(self, request):
         handler = self.default_auth_grant
 
-        if "openid" in request.scopes:
+        if request.scopes and "openid" in request.scopes:
             handler = self.oidc_auth_grant
 
         log.debug('Selecting handler for request %r.', handler)
@@ -77,10 +76,92 @@ class AuthCodeGrantDispatcher(object):
         return self._handler_for_request(request).validate_authorization_request(request)
 
 
-class OpenIDConnectBase(GrantTypeBase):
+class ImplicitTokenGrantDispatcher(object):
+    """
+    This is an adapter class that will route simple Authorization Code requests, those that have response_type=code and a scope
+    including 'openid' to either the default_auth_grant or the oidc_auth_grant based on the scopes requested.
+    """
+    def __init__(self, default_implicit_grant=None, oidc_implicit_grant=None):
+        self.default_implicit_grant = default_implicit_grant
+        self.oidc_implicit_grant = oidc_implicit_grant
 
-    def __init__(self, request_validator=None):
-        self.request_validator = request_validator or RequestValidator()
+    def _handler_for_request(self, request):
+        handler = self.default_implicit_grant
+
+        if request.scopes and "openid" in request.scopes and 'id_token' in request.response_type:
+            handler = self.oidc_implicit_grant
+
+        log.debug('Selecting handler for request %r.', handler)
+        return handler
+
+    def create_authorization_response(self, request, token_handler):
+        return self._handler_for_request(request).create_authorization_response(request, token_handler)
+
+    def validate_authorization_request(self, request):
+        return self._handler_for_request(request).validate_authorization_request(request)
+
+
+class AuthTokenGrantDispatcher(object):
+    """
+    This is an adapter class that will route simple Token requests, those that authorization_code have a scope
+    including 'openid' to either the default_token_grant or the oidc_token_grant based on the scopes requested.
+    """
+    def __init__(self, request_validator, default_token_grant=None, oidc_token_grant=None):
+        self.default_token_grant = default_token_grant
+        self.oidc_token_grant = oidc_token_grant
+        self.request_validator = request_validator
+
+    def _handler_for_request(self, request):
+        handler = self.default_token_grant
+        scopes = ()
+        parameters = dict(request.decoded_body)
+        client_id = parameters.get('client_id', None)
+        code = parameters.get('code', None)
+        redirect_uri = parameters.get('redirect_uri', None)
+
+        # If code is not pressent fallback to `default_token_grant` wich will
+        # raise an error for the missing `code` in `create_token_response` step.
+        if code:
+            scopes = self.request_validator.get_authorization_code_scopes(client_id, code, redirect_uri, request)
+
+        if 'openid' in scopes:
+            handler = self.oidc_token_grant
+
+        log.debug('Selecting handler for request %r.', handler)
+        return handler
+
+    def create_token_response(self, request, token_handler):
+        handler = self._handler_for_request(request)
+        return handler.create_token_response(request, token_handler)
+
+
+class OpenIDConnectBase(object):
+
+    # Just proxy the majority of method calls through to the
+    # proxy_target grant type handler, which will usually be either
+    # the standard OAuth2 AuthCode or Implicit grant types.
+    def __getattr__(self, attr):
+        return getattr(self.proxy_target, attr)
+
+    def __setattr__(self, attr, value):
+        proxied_attrs = set(('refresh_token', 'response_types'))
+        if attr in proxied_attrs:
+            setattr(self.proxy_target, attr, value)
+        else:
+            super(OpenIDConnectBase, self).__setattr__(attr, value)
+
+    def validate_authorization_request(self, request):
+        """Validates the OpenID Connect authorization request parameters.
+
+        :returns: (list of scopes, dict of request info)
+        """
+        # If request.prompt is 'none' then no login/authorization form should
+        # be presented to the user. Instead, a silent login/authorization
+        # should be performed.
+        if request.prompt == 'none':
+            raise OIDCNoPrompt()
+        else:
+            return self.proxy_target.validate_authorization_request(request)
 
     def _inflate_claims(self, request):
         # this may be called multiple times in a single request so make sure we only de-serialize the claims once
@@ -96,7 +177,7 @@ class OpenIDConnectBase(GrantTypeBase):
 
     def add_id_token(self, token, token_handler, request):
         # Treat it as normal OAuth 2 auth code request if openid is not present
-        if 'openid' not in request.scopes:
+        if not request.scopes or 'openid' not in request.scopes:
             return token
 
         # Only add an id token on auth/token step if asked for.
@@ -119,6 +200,13 @@ class OpenIDConnectBase(GrantTypeBase):
     def openid_authorization_validator(self, request):
         """Perform OpenID Connect specific authorization request validation.
 
+        nonce
+                OPTIONAL. String value used to associate a Client session with
+                an ID Token, and to mitigate replay attacks. The value is
+                passed through unmodified from the Authentication Request to
+                the ID Token. Sufficient entropy MUST be present in the nonce
+                values used to prevent attackers from guessing values
+
         display
                 OPTIONAL. ASCII string value that specifies how the
                 Authorization Server displays the authentication and consent
@@ -249,31 +337,43 @@ class OpenIDConnectBase(GrantTypeBase):
         """
 
         # Treat it as normal OAuth 2 auth code request if openid is not present
-        if not 'openid' in request.scopes:
+        if not request.scopes or 'openid' not in request.scopes:
             return {}
 
-        # prompt other than 'none' should be handled by the server code that uses oauthlib
-        if request.prompt == 'none' and not request.id_token_hint:
-            msg = "Prompt is set to none yet id_token_hint is missing."
-            raise InvalidRequestError(request=request, description=msg)
+        prompt = request.prompt if request.prompt else []
+        if hasattr(prompt, 'split'):
+            prompt = prompt.strip().split()
+        prompt = set(prompt)
+
+        if 'none' in prompt:
+
+            if len(prompt) > 1:
+                msg = "Prompt none is mutually exclusive with other values."
+                raise InvalidRequestError(request=request, description=msg)
+
+            # prompt other than 'none' should be handled by the server code that
+            # uses oauthlib
+            if not request.id_token_hint:
+                msg = "Prompt is set to none yet id_token_hint is missing."
+                raise InvalidRequestError(request=request, description=msg)
 
-        if request.prompt == 'none':
             if not self.request_validator.validate_silent_login(request):
                 raise LoginRequired(request=request)
+
             if not self.request_validator.validate_silent_authorization(request):
                 raise ConsentRequired(request=request)
 
         self._inflate_claims(request)
 
         if not self.request_validator.validate_user_match(
-            request.id_token_hint, request.scopes, request.claims, request):
+                request.id_token_hint, request.scopes, request.claims, request):
             msg = "Session user does not match client supplied user."
             raise LoginRequired(request=request, description=msg)
 
-
         request_info = {
             'display': request.display,
-            'prompt': request.prompt.split() if request.prompt else [],
+            'nonce': request.nonce,
+            'prompt': prompt,
             'ui_locales': request.ui_locales.split() if request.ui_locales else [],
             'id_token_hint': request.id_token_hint,
             'login_hint': request.login_hint,
@@ -289,7 +389,8 @@ class OpenIDConnectBase(GrantTypeBase):
         if request.response_type == 'token':
             return {}
 
-        if not 'openid' in request.scopes:
+        # Treat it as normal OAuth 2 auth code request if openid is not present
+        if not request.scopes or 'openid' not in request.scopes:
             return {}
 
         # REQUIRED. String value used to associate a Client session with an ID
@@ -302,141 +403,49 @@ class OpenIDConnectBase(GrantTypeBase):
             desc = 'Request is missing mandatory nonce parameter.'
             raise InvalidRequestError(request=request, description=desc)
 
-        self._inflate_claims(request)
+        return {}
 
-        return {'nonce': request.nonce, 'claims': request.claims}
 
 class OpenIDConnectAuthCode(OpenIDConnectBase):
 
-    def __init__(self, request_validator=None):
-        self.request_validator = request_validator or RequestValidator()
-        super(OpenIDConnectAuthCode, self).__init__(
-            request_validator=self.request_validator)
-        self.auth_code = AuthorizationCodeGrant(
-            request_validator=self.request_validator)
-        self.auth_code.register_authorization_validator(
+    def __init__(self, request_validator=None, **kwargs):
+        self.proxy_target = AuthorizationCodeGrant(
+            request_validator=request_validator, **kwargs)
+        self.custom_validators.post_auth.append(
             self.openid_authorization_validator)
-        self.auth_code.register_token_modifier(self.add_id_token)
-
-    @property
-    def refresh_token(self):
-        return self.auth_code.refresh_token
-
-    @refresh_token.setter
-    def refresh_token(self, value):
-        self.auth_code.refresh_token = value
-
-    def create_authorization_code(self, request):
-        return self.auth_code.create_authorization_code(request)
-
-    def create_authorization_response(self, request, token_handler):
-        return self.auth_code.create_authorization_response(
-            request, token_handler)
-
-    def create_token_response(self, request, token_handler):
-        return self.auth_code.create_token_response(request, token_handler)
-
-    def validate_authorization_request(self, request):
-        """Validates the OpenID Connect authorization request parameters.
-
-        :returns: (list of scopes, dict of request info)
-        """
-        # If request.prompt is 'none' then no login/authorization form should
-        # be presented to the user. Instead, a silent login/authorization
-        # should be performed.
-        if request.prompt == 'none':
-            raise OIDCNoPrompt()
-        else:
-            return self.auth_code.validate_authorization_request(request)
-
-    def validate_token_request(self, request):
-        return self.auth_code.validate_token_request(request)
+        self.register_token_modifier(self.add_id_token)
 
 
 class OpenIDConnectImplicit(OpenIDConnectBase):
 
-    def __init__(self, request_validator=None):
-        self.request_validator = request_validator or RequestValidator()
-        super(OpenIDConnectImplicit, self).__init__(
-            request_validator=self.request_validator)
-        self.implicit = ImplicitGrant(
-            request_validator=request_validator)
-        self.implicit.register_response_type('id_token')
-        self.implicit.register_response_type('id_token token')
-        self.implicit.register_authorization_validator(
+    def __init__(self, request_validator=None, **kwargs):
+        self.proxy_target = ImplicitGrant(
+            request_validator=request_validator, **kwargs)
+        self.register_response_type('id_token')
+        self.register_response_type('id_token token')
+        self.custom_validators.post_auth.append(
             self.openid_authorization_validator)
-        self.implicit.register_authorization_validator(
+        self.custom_validators.post_auth.append(
             self.openid_implicit_authorization_validator)
-        self.implicit.register_token_modifier(self.add_id_token)
-
-    def create_authorization_response(self, request, token_handler):
-        return self.create_token_response(request, token_handler)
-
-    def create_token_response(self, request, token_handler):
-        return self.implicit.create_authorization_response(
-            request, token_handler)
-
-    def validate_authorization_request(self, request):
-        """Validates the OpenID Connect authorization request parameters.
-
-        :returns: (list of scopes, dict of request info)
-        """
-        # If request.prompt is 'none' then no login/authorization form should
-        # be presented to the user. Instead, a silent login/authorization
-        # should be performed.
-        if request.prompt == 'none':
-            raise OIDCNoPrompt()
-        else:
-            return self.implicit.validate_authorization_request(request)
+        self.register_token_modifier(self.add_id_token)
 
 
 class OpenIDConnectHybrid(OpenIDConnectBase):
 
-    def __init__(self, request_validator=None):
+    def __init__(self, request_validator=None, **kwargs):
         self.request_validator = request_validator or RequestValidator()
 
-        self.auth_code = AuthorizationCodeGrant(
-            request_validator=request_validator)
-        self.auth_code.register_response_type('code id_token')
-        self.auth_code.register_response_type('code token')
-        self.auth_code.register_response_type('code id_token token')
-        self.auth_code.register_authorization_validator(
+        self.proxy_target = AuthorizationCodeGrant(
+            request_validator=request_validator, **kwargs)
+        # All hybrid response types should be fragment-encoded.
+        self.proxy_target.default_response_mode = "fragment"
+        self.register_response_type('code id_token')
+        self.register_response_type('code token')
+        self.register_response_type('code id_token token')
+        self.custom_validators.post_auth.append(
             self.openid_authorization_validator)
-        self.auth_code.register_code_modifier(self.add_token)
-        self.auth_code.register_code_modifier(self.add_id_token)
-        self.auth_code.register_token_modifier(self.add_id_token)
-
-    @property
-    def refresh_token(self):
-        return self.auth_code.refresh_token
-
-    @refresh_token.setter
-    def refresh_token(self, value):
-        self.auth_code.refresh_token = value
-
-    def create_authorization_code(self, request):
-        return self.auth_code.create_authorization_code(request)
-
-    def create_authorization_response(self, request, token_handler):
-        return self.auth_code.create_authorization_response(
-            request, token_handler)
-
-    def create_token_response(self, request, token_handler):
-        return self.auth_code.create_token_response(request, token_handler)
-
-    def validate_authorization_request(self, request):
-        """Validates the OpenID Connect authorization request parameters.
-
-        :returns: (list of scopes, dict of request info)
-        """
-        # If request.prompt is 'none' then no login/authorization form should
-        # be presented to the user. Instead, a silent login/authorization
-        # should be performed.
-        if request.prompt == 'none':
-            raise OIDCNoPrompt()
-        else:
-            return self.auth_code.validate_authorization_request(request)
-
-    def validate_token_request(self, request):
-        return self.auth_code.validate_token_request(request)
-
+        # Hybrid flows can return the id_token from the authorization
+        # endpoint as part of the 'code' response
+        self.register_code_modifier(self.add_token)
+        self.register_code_modifier(self.add_id_token)
+        self.register_token_modifier(self.add_id_token)
